open Core

module Expression(N : Number.NUMBER) = struct
    type number = N.number

    type variable = string
    exception UndefinedVariable of string 

    type binary_op = Plus | Times | Minus | Divide | Power
    type comparison = Equal | LessThan | NotEqual (*| LessThan  | GreaterThan |  GreaterThanOrEqual | LessThanOrEqual | NotEqual | In | NotIn | Is | IsNot*)
    type unary_op = Negation
    type expression = 
        | Var of variable 
        | Num of number 
        | List of expression list
        | Boolean of bool
        | BinOp of binary_op * expression * expression
        | UnaryOp of unary_op * expression
        | ListComprehension of expression * expression
        | ComprehensionFor of variable * expression * expression
        | ComprehensionIf of expression * expression
        | End
        | IfElse of expression * expression * expression 
        | Or of expression * expression 
        | And of expression * expression 
        | Not of expression 
        | Comparison of comparison * expression * expression
	and statement =
		| Assign of variable * expression
		| Print of expression
	and program = 
		| Expression of expression 
		| Statement of statement 

    type primitive_type =
        | NUMBER
        | BOOLEAN
        | LIST

    let string_of_primitive_type name = function 
        | NUMBER -> "int " ^ name ^ ";"
        | BOOLEAN -> "boolean " ^ name ^ ";"
        | LIST -> "List " ^ name ^ " = new LinkedList();"

    let rec string_of_expression_list l : string = let rec f = function 
        | [] -> ""
        | [a] -> string_of_expression a 
        | x::xs -> (string_of_expression x) ^ ", " ^ (f xs)
    in "[" ^ f l ^ "]"
    and string_of_expression : expression -> string = function
        | (Num n) -> N.string_of_number n
        | (Var v) -> v
        | (Boolean b) -> string_of_bool b
        | (List l) ->  string_of_expression_list l
        | BinOp(Plus, e1, e2) -> "(" ^ string_of_expression e1 ^ "+" ^ string_of_expression e2 ^ ")"
        | BinOp(Times, e1, e2) -> "(" ^ string_of_expression e1 ^ "*" ^ string_of_expression e2 ^ ")"
        | BinOp(Divide, e1, e2) -> "(" ^ string_of_expression e1 ^ "/" ^ string_of_expression e2 ^ ")"
        | BinOp(Minus, e1, e2) -> "(" ^ string_of_expression e1 ^ "-" ^ string_of_expression e2 ^ ")"
        | BinOp(Power, e1, e2) -> string_of_expression e1 ^ "^(" ^ string_of_expression e2 ^ ")"
        | UnaryOp(Negation, e) -> "-" ^ string_of_expression e
        | IfElse(t, e1, e2) -> string_of_expression e1 ^ " if " ^ string_of_expression t ^ " else " ^ string_of_expression e2
        | And(t1, t2) -> string_of_expression t1 ^ " and " ^ string_of_expression t2
        | Or(t1, t2) -> string_of_expression t1 ^ " or " ^ string_of_expression t2
        | Not(t) -> "not " ^ string_of_expression t
        | Comparison(Equal, e1, e2) -> string_of_expression e1 ^ " == " ^ string_of_expression e2
        | Comparison(LessThan, e1, e2) -> string_of_expression e1 ^ " < " ^ string_of_expression e2
        | Comparison(NotEqual, e1, e2) -> string_of_expression e1 ^ " != " ^ string_of_expression e2
        | ListComprehension(expr, comprehension) -> "[" ^ string_of_expression expr ^ string_of_expression comprehension ^ "]"
        | ComprehensionFor(var, lst, next) -> " for " ^ string_of_expression (Var var) ^ " in " ^ string_of_expression lst ^ string_of_expression next 
        | ComprehensionIf(var, next) -> " if " ^ string_of_expression var ^ string_of_expression next
        | End -> "" 

	exception EvaluatedAssign of string
    exception TypeError of string 
    exception SyntaxError of string 

    let unwrap_number e = match e with
        | Num n -> n
        | Boolean b -> if b then N.one else N.zero
        | _ -> raise (TypeError ("Expected <number>, got: " ^ (string_of_expression e)))

    let unwrap_boolean e = match e with
        | Boolean b -> b
        | Num n -> if n = N.zero then false else true
        | _ -> raise (TypeError ("Expected <boolean>, got: " ^ (string_of_expression e)))

    let unwrap_list e = match e with
        | List l -> l
        | _ -> raise (TypeError ("Expected <list>, got: " ^ (string_of_expression e)))


    let rec eval_expression vars : expression -> expression = function 
        | (Num n) -> Num n
        | (Boolean b) -> Boolean b
        | (Var v) -> (match Map.find vars v with
           | Some n -> n
           | None -> raise (UndefinedVariable v)
          )
        | (List l) -> List (List.map l ~f:(fun e -> eval_expression vars e))
        | BinOp(Plus, e1, e2) -> (
            let e1 = eval_expression vars e1 in
            let e2 = eval_expression vars e2 in
                try
                    Num (N.add (unwrap_number e1) (unwrap_number e2))
                with
                    TypeError _ -> try List (unwrap_list e1 @ (unwrap_list e2))
                    with TypeError _ -> raise (TypeError "Expected <number,number> or <list,list>")
        )
        | BinOp(Times, e1, e2) -> Num (N.multiply(unwrap_number (eval_expression vars e1)) (unwrap_number (eval_expression vars e2)))
        | BinOp(Divide, e1, e2) -> Num (N.divide (unwrap_number (eval_expression vars e1)) (unwrap_number (eval_expression vars e2)))
        | BinOp(Minus, e1, e2) -> Num (N.subtract (unwrap_number (eval_expression vars e1)) (unwrap_number (eval_expression vars e2)))
        | BinOp(Power, e1, e2) -> Num (N.pow (unwrap_number (eval_expression vars e1)) (unwrap_number (eval_expression vars e2)))
        | UnaryOp(Negation, e) -> Num (N.negative (unwrap_number (eval_expression vars e)))
        | IfElse(t, e1, e2)    -> (match unwrap_boolean (eval_expression vars t) with
                                    | true -> eval_expression vars e1
                                    | false -> eval_expression vars e2
                                )
        | And(t1, t2) -> (match unwrap_boolean (eval_expression vars t1) with
                            | true -> (eval_expression vars t2)
                            | false -> Boolean false
                        )
        | Or(t1, t2) -> (match unwrap_boolean (eval_expression vars t1) with
                            | false -> (eval_expression vars t2)
                            | true -> Boolean true 
                        )
        | Not(t) -> (match (eval_expression vars t) with
                        | Boolean false -> Boolean true
                        | Boolean true -> Boolean false 
                        | e -> raise (TypeError ("Expected <boolean>, got: " ^ (string_of_expression e)))
                    )
        | Comparison(Equal, e1, e2) -> let r1 = eval_expression vars e1 in
                                       let r2 = eval_expression vars e2 in
                                         Boolean (r1 = r2)
        | Comparison(LessThan, e1, e2) -> let r1 = eval_expression vars e1 in
                                       let r2 = eval_expression vars e2 in
                                         Boolean (r1 < r2)
        | Comparison(NotEqual, e1, e2) -> let r1 = eval_expression vars e1 in
                                       let r2 = eval_expression vars e2 in
                                         Boolean (r1 <> r2)

        | ListComprehension(expr, comprehension) -> (
                let rec recurse = function 
                    | End -> expr 
                    | ComprehensionFor(var, lst, next) -> ComprehensionFor(var, lst, recurse next)
                    | ComprehensionIf(expr_in, next) -> ComprehensionIf(expr_in, recurse next)
                    | e -> raise (SyntaxError ("Expected <comprehension>, got: " ^ (string_of_expression e)))
                in
                    eval_expression vars (recurse comprehension)
        )
        | ComprehensionFor(var, lst, next) -> (
            let lst = eval_expression vars lst in
                let rec loop = function 
                    | [] -> []
                    | x::l -> (
                        let vars = Map.set vars ~key:var ~data:(eval_expression vars x) in
                        (let evaled = eval_expression vars next in (
                            match evaled with
                                | End -> loop l
                                | List e -> (match next with 
                                    | ComprehensionFor(_, _, _) -> e @ (loop l)
                                    | _ -> (List e)::(loop l)
                                )
                                | e -> e::(loop l)
                        ))
                    )
                in match lst with
                    | List l -> List (loop l)
                    | e -> raise (TypeError ("Expected <list>, got: " ^ (string_of_expression e)))
        )
        | ComprehensionIf(expr, next) -> (
            let if_expr = eval_expression vars expr in (
                match if_expr with
                    | Boolean false -> End
                    | Boolean true -> eval_expression vars next
                    | e -> raise (TypeError ("Expected <boolean>, got: " ^ (string_of_expression e)))
            )
        )
        | End -> raise (SyntaxError ("Expected <comprehension>, got: End"))

    let rec type_expression vars : expression -> primitive_type = function 
        | (Num _) -> NUMBER 
        | (Boolean _) -> BOOLEAN 
        | (Var v) -> (match Map.find vars v with
           | Some n -> n
           | None -> raise (UndefinedVariable v)
          )
        | (List _) -> LIST 
        | BinOp(Plus, e1, e2) -> (
            let t1 = type_expression vars e1 in
            let t2 = type_expression vars e2 in
                if t1 = t2 then t1 else raise (TypeError "Expected <T,T>")
        )
        | BinOp(_, _, _) -> NUMBER
        | UnaryOp(Negation, _) -> NUMBER 
        | IfElse(_, e1, e2)    -> (
            let t1 = type_expression vars e1 in
            let t2 = type_expression vars e2 in
                if t1 = t2 then t1 else raise (TypeError "Expected <T,T>")
        )
        | And(_, _) -> BOOLEAN
        | Or(_, _) -> BOOLEAN
        | Not(_) -> BOOLEAN
        | Comparison(_, _, _) -> BOOLEAN
        | ListComprehension(_, _) -> LIST
        | ComprehensionFor(_, _, _) -> raise (TypeError ("Unexpected comprehension for")) 
        | ComprehensionIf(_, _) -> raise (TypeError ("Unexpected comprehension if")) 
        | End -> raise (TypeError ("Unexpected end"))


    let emit_declaration v type_env expr = 
        Out_channel.output_string stdout "        ";
        Out_channel.output_string stdout (string_of_primitive_type v (type_expression type_env expr));
        Out_channel.newline stdout

    let emit_assignment v str = 
        Out_channel.output_string stdout "        ";
        Out_channel.output_string stdout ( v ^ " = " ^ str ^ ";");
        Out_channel.newline stdout

    let make_temp_var var_idx = "_" ^ (string_of_int var_idx)

    let rec emit_expression vars idx: expression -> variable * int = function 
        | (Num n) -> raise (TypeError "Not implemented")
        | (Boolean b) -> raise (TypeError "Not implemented")
        | (Var v) -> raise (TypeError "Not implemented")
        | (List l) -> raise (TypeError "Not implemented")
        | BinOp(Plus, e1, e2) -> raise (TypeError "Not implemented")
        | BinOp(Times, e1, e2) -> raise (TypeError "Not implemented")
        | BinOp(Divide, e1, e2) -> raise (TypeError "Not implemented")
        | BinOp(Minus, e1, e2) -> raise (TypeError "Not implemented")
        | BinOp(Power, e1, e2) -> raise (TypeError "Not implemented")
        | UnaryOp(Negation, e) -> raise (TypeError "Not implemented")
        | IfElse(t, e1, e2) -> raise (TypeError "Not implemented")
        | And(t1, t2) -> raise (TypeError "Not implemented")
        | Or(t1, t2) -> raise (TypeError "Not implemented")
        | Not(t) -> raise (TypeError "Not implemented")
        | Comparison(Equal, e1, e2) -> raise (TypeError "Not implemented")
        | Comparison(LessThan, e1, e2) -> raise (TypeError "Not implemented")
        | Comparison(NotEqual, e1, e2) -> raise (TypeError "Not implemented")
        | ListComprehension(expr, comprehension) -> raise (TypeError "Not implemented")
        | ComprehensionFor(var, lst, next) -> raise (TypeError "Not implemented")
        | ComprehensionIf(expr, next) -> raise (TypeError "Not implemented")
        | End -> raise (SyntaxError ("Expected <comprehension>, got: End"))
end
